package src.hw7S10;

import java.util.*;

public class PackingSolver {
	private int capacity;    // capacity of truck
	private int[] items;  // input weight of items
	private int[] groups; // the number of items of each size
	private int max_case; // max size of the items
	private ArrayList<Pattern> solutions;
	
	public PackingSolver(int w, int[] c)
	{
		capacity = w;
		items = c;

		// groups[i] stores the number of items of size i;
		groups = new int[capacity+1];
		for (int i=0, e=items.length; i<e; ++i )
		{
			assert( items[i] > 0 );
			if ( items[i] <= capacity )
				++groups[items[i]];
		}
		
		// set max_item to the maximal size of the items
		for ( max_case=capacity; max_case>=0; --max_case )
			if ( groups[max_case] > 0 ) break;
		
		solutions = new ArrayList<Pattern>(); 
	}

	public void solve()
	{
		boolean use_preprocessing = true, show_preprocessing = false;
		if (use_preprocessing)
		{
			preprocessing();
			if ( show_preprocessing )
				print_solutions();
		}
		sum_of_subset_bb();
		print_solutions();
	}
	
	// heuristics for getting "obvious" patterns
	private void preprocessing()
	{
		// Although we could do preprocessing for all items, 
		// it tends to use the small items "unwisely", 
		// which may limit the solution space that branch&bound can explore 
		// and resulting worse overall results.  
		final int min_size_to_preprocess = max_case/2;
		
		for(int i=max_case; i>min_size_to_preprocess; --i)
		{
			// if truck capacity is a multiple of item size and we have enough items
			int count_i = capacity / i;
			if ((capacity % i == 0) && (count_i <= groups[i]) )
			{
				int count_p = groups[i] / count_i;

				ArrayList<Pair> ap = new ArrayList<Pair>();
				ap.add(new Pair(i, count_i));
				Pattern pt = new Pattern(ap, count_p);
				
				solutions.add(pt);
				// take away the used items
				groups[i] -= count_p * count_i;
			}
			else // try a greedy heuristics
			{
				ArrayList<Pair> ap = new ArrayList<Pair>();
				int remain= capacity;
				// put bigger size items first
				for ( int j=i; j>0 && remain > 0; --j)
				{
					int count_j = remain / j;
					if ( count_j > groups[j] )
						count_j = groups[j];
					if ( count_j > 0 )
					{
						ap.add(new Pair(j, count_j));
						remain -= count_j * j;
					}
				}
				
				// If a pattern is complete, add it to solutions
				if ( remain == 0 ) 
				{
					assert( !ap.isEmpty() );
					
					// Find out how many replications of this pattern can be allocated.
					int count_p = Integer.MAX_VALUE; // a upper bound
					for( Pair p: ap)
					{
						count_p = Math.min(count_p, groups[p.first] / p.second);
					}
					
					Pattern pt = new Pattern(ap, count_p);
					solutions.add(pt);
					
					// take away the used items
					for( Pair p: ap)
						groups[p.first] -= p.second * count_p;
				}
			}
		}
	}
	
	private void print_solutions()
	{
		int count_truck = 0, count_full = 0, size_unfilled = 0;
		for( int i=0, e=solutions.size(); i<e; ++i)
		{
			Pattern pt = solutions.get(i);
			boolean full = pt.checksum(capacity);
			System.out.println("Pattern " + i + ": " + pt.toString() + ". Full? " + full);
			count_truck += pt.allocated;
			if ( full ) 
				count_full += pt.allocated;
			else
				size_unfilled += pt.allocated * (capacity - pt.width());
		}
		
		System.out.println("Total allocated: " + count_truck + "\tFull: " + count_full + "\tPercent: " + (double)(count_full)/count_truck );
		
		System.out.println("======================");
		System.out.println("Incomplete (by group): ");
		int remain_cases = 0;
		for(int i=1; i<=max_case; ++i)
		{
			if ( groups[i] > 0) {
				System.out.println("size " + i + "x" + groups[i]);
				remain_cases += groups[i];
			}
		}
		System.out.println("Total remain: " + remain_cases);
		
		if ( remain_cases == 0 )
		{
			System.out.println("Allocation complete. " );
			System.out.println("Fully filled: " + count_full + "\tPartially filled: " + (count_truck-count_full)
					+ "\tTotal pallets: " + count_truck + "\tPercent space filled: " + (double)(count_truck*capacity-size_unfilled)/(count_truck*capacity) );			
		}
	}
	
	private void sum_of_subset_bb() {
		SumOfSubsetBB sosbb = new SumOfSubsetBB(groups, capacity);
		ArrayList<Pattern> bbsol = sosbb.solve();
		solutions.addAll(bbsol);
	}
}