package src.hw7S10;

import java.util.*;

public class SumOfSubsetBB {
	private int width;
	private int[] groups;
	private int[] sorted_cases;
	private int[] selected;
	private int sum_remained;
	private Stack<BBNode> discovered; 
	private ArrayList<Pattern> solutions;
	private boolean show_debug;
	
	public SumOfSubsetBB(int[] groups, int width)
	{
		this.width = width;
		this.groups = groups;
		discovered = new Stack<BBNode>();
		solutions = new ArrayList<Pattern>();
		
		this.show_debug = false; // do not print out debug message
		init();
	}

	private void init()
	{
		sum_remained = 0;
		int count = 0, count_groups = groups.length;
		for (int i=1; i<count_groups; ++i)
		{
			count += groups[i];
			sum_remained += groups[i] * i;
		}
		if (show_debug)
			System.out.println("Count=" + count + ", sum_remained=" + sum_remained);
		sorted_cases = new int[count];
		
		// sort descending might be better (sort ascending works, too)
		for (int i=count_groups-1, j=0; i>0; --i )
		{
			// fill i for groups[i] times
			for( int k=0; k<groups[i]; ++k)
				sorted_cases[j++] = i;	
		}
		
		selected = new int[count]; // initialized to 0
		
		discovered.clear();
	}
	
	// a set of return codes
	final private int RET_FINISH = 0;      	// no more case remained
	final private int RET_INCOMPLETE = -1; 	// an incomplete pattern is returned (and nothing left)
	final private int RET_COMPLETE = 1;		// a complete pattern is returned 


	public ArrayList<Pattern> solve()
	{
		for (int ret_code = sumofsubset(); ret_code != RET_FINISH; ret_code = sumofsubset())
		{
			switch(ret_code)
			{
			case RET_INCOMPLETE	: // create pattern with all remained cases
			{
				ArrayList<Pair> ap = new ArrayList<Pair>();
				int sum = 0; // verifies if the pattern is really incomplete 
				for( int i=groups.length-1; i>0; --i)
				{
					if ( groups[i] > 0 ) {
						ap.add( new Pair(i, groups[i]));
						sum += i * groups[i];
					}
				}
				assert( sum < width );

				Pattern pt = new Pattern(ap, 1);

				solutions.add(pt);
				// take away the used cases
				for( Pair p: ap)
					groups[p.first] -= p.second;
				
				if ( show_debug )
				{
					boolean d = done(); 
					System.out.println("Done? " + d); 
					//assert( d );
				}
				
				return solutions;
				//break;
			}	
			case RET_COMPLETE :
			{
				ArrayList<Pair> ap = new ArrayList<Pair>();

				// count selected cases and put them into pairs. (this code assumes "reference" type is used in ArrayList)
				int last_added = width+1;
				Pair newpair = null;

				for (int i=0; i<selected.length; ++i)
				{
					if ( selected[i] == 1 )
					{
						if ( sorted_cases[i] == last_added )
							newpair.second++;
						else
						{
							last_added = sorted_cases[i];
							newpair = new Pair(last_added, 1);
							ap.add(newpair);
						}
					}
				}
				
				// verification
				int sum = 0; 
				for( Pair p: ap )
				{
					sum += p.first * p.second;
					if ( show_debug )
						System.out.println(p.first + "x" + p.second + " ");
				}
				assert( sum == width );

				// Find out how many replications of this pattern can be allocated.
				int count_p = Integer.MAX_VALUE; // a upper bound
				for( Pair p: ap)
				{
					count_p = Math.min(count_p, groups[p.first] / p.second);
				}

				int ap_size = ap.size();
				if ( show_debug )
					System.out.println("ap_size=" + ap_size);
				assert( ap_size > 0 );
				
				//int maxc = ap.get(0).first, minc = ap.get(ap_size-1).first;
				Pattern pt = new Pattern(ap, count_p);
				solutions.add(pt);

				// take away the used cases
				for( Pair p: ap)
					groups[p.first] -= p.second * count_p;

				// re-initialize (for next branch&bound run)
				init();
			}
			break;
			}
		}
		
		if ( show_debug )
		{
			boolean d = done(); 
			System.out.println("Done? " + d); 
			//assert( d );
		}
		
		return solutions;
	}

	// check if any case is left
	private boolean done()
	{
		for (int i=1; i<groups.length; ++i)
			if ( groups[i] > 0 )
				return false;
		
		return true;
	}

	private int sumofsubset()
	{
		if (show_debug)
			System.out.println("sum_remained=" + sum_remained + ", width=" + width);
		
		if ( sum_remained == 0 )
			return RET_FINISH; // not necessary
		
		if ( sum_remained < width )
			return RET_INCOMPLETE; // an incomplete pattern
		
		int n = sorted_cases.length;
		int k = 0, sum = 0 ; // current item; sum-so-far (before considering k)
		do {
			while ( k < n )
			{
				int next_sum = sum + sorted_cases[k];
				if( next_sum == width )
				{
					selected[k] = 1;
					return RET_COMPLETE; // a complete pattern is found
				}
				
				if( next_sum < width ) 
				{
					sum_remained -= sorted_cases[k];
					// store current state for backtracking, if right child is feasible
					if ( sum + sum_remained >= width &&  k+1 < n && sum + sorted_cases[k+1] <= width )
						discovered.add(new BBNode(sum, selected, k, sum_remained));

					selected[k] = 1;  // create left child
					sum = next_sum;
					
					++k;
				}
				else // next_sum > width
					break; // no more left child allowed; need backtrack
			}

			// either k == n or no more left child is allowed
			if ( k == n )
				return RET_INCOMPLETE; // incomplete pattern
			
			if ( discovered.empty() )
				return RET_FINISH; // backtrack completed; no solution found
			
			// backtrack
			BBNode pop_node = discovered.pop();
			// create right child
			sum = pop_node.sum;
			sum_remained = pop_node.remain;
			k = pop_node.x.length;
			for ( int i=0; i<k; ++i)
				selected[i] = pop_node.x[i];
			selected[k] = 0;
			++k;
		} while (true);
	}
}
